indexing

   title:       "Integer commercial formatter class";
   cluster:     "format/commerce [format/commerce]";
   project:     "The Universal Simple Formatting Library (USFL)";
   copyright:   "SIG Computer, 1995";
   author:      "Igor Nesterof (nest@eiffel.ru)";
   original:    08,Sep,1995;
   version:     1.0;
   done_at:     "SIG Computer (info@eiffel.ru)";
   extrnl_name: "int_com.e"

class INTEGER_COMMERCIAL_FORMATTER

inherit
   ABSTRACT_COMMERCIAL_FORMATTER
      redefine make, output, set_sign_leading, set_blank_when_zero,
               set_sign_leading_separate, set_sign_trailing, set_currency, set_delimiters
   end
   EXTENDED_MATH
   STRING_TOOLS

creation
   make

feature

   make (i : INTEGER_REF) is
      do
         sign := int_sign(i.item)
         num := int_abs(i.item)
         blank_when_zero := False
         sign_leading := False
         fill_char := ' '
         currency := '$'
         delimiters := ""
      end

   set_blank_when_zero is
      do
         blank_when_zero := True
      end

--   set_justified_right is
--   set_justified_left is

   set_sign_leading is
      do
         sign_leading := True
      end

   set_sign_leading_separate is
      local
         s : string
      do
         if contain(picture,'+') then
            raise_exception("No more sign")
         end
         !!s.make(0)
         x_append(s,picture)
         s.precede('+')
         set_picture(s)
      end

   set_sign_trailing is
      local
         s : string
      do
         if contain(picture,'+') then
            raise_exception("No more sign")
         end
         !!s.make(0)
         x_append(s,picture)
         x_extend(s,'+')
         set_picture(s)
      end

   set_currency(s : STRING) is
   require else
      String_Not_Void : s /= Void
      do
         currency := s.item(1)
      end

   set_delimiters(s : STRING) is
      do
         x_adapt(delimiters,s)
      end

   output : STRING is
   require else
      Non_void_picture : picture /= Void
      do
         from
            init_output
         until
            cur_pic.is_equal("")
         loop
            next_char
         end
         check_output
         Result := cur_str
      end

feature {INTEGER_COMMERCIAL_FORMATTER}

   -- all attributes with prefix cur_ is current state of parsing
   -- parsing comes from last char to first
   cur_pic, cur_str, delimiters : STRING
   num, cur_num, sign : INTEGER

   is_last_char : BOOLEAN
   is_yet_sign : BOOLEAN
   blank_when_zero : BOOLEAN
   sign_leading : BOOLEAN
   fill_char    : CHARACTER
   currency : CHARACTER

   digits : STRING is "0123456789"

   init_output is
      do
         table := << False, False, False, False >>
         cur_num := num
         cur_pic := picture
         cur_str := ""
         is_last_char := true
         is_yet_sign := false
         delimiters := "_/:,"
      end

   check_output is
      do
         if cur_num /= 0 then
            raise_exception("Number not fit")
         end
--         if sign = -1 and not is_yet_sign then
--            raise_exception("No sign place")
--         end
      end

   next_char is
      local
         pat, data_char : CHARACTER
      do
         pat := cur_pic.item(cur_pic.count)
         if pat = '9' then
            check_digit
            data_char  := produce_digit('9')
         elseif contain("Z*",pat) then
            check_blank
            data_char  := produce_blank(pat)
         elseif contain("+-",pat) then
            check_sign
            data_char  := produce_sign(pat)
         elseif pat = '$' then
            check_dollar
            data_char := produce_dollar
         elseif contain(delimiters,pat) then
            data_char := produce_noise(pat)
         elseif contain("B",pat) then
            data_char := produce_noise(pat)
         else
            raise_exception("Illegal picture char")
         end
         cur_str.precede(data_char)
         cur_pic.remove(cur_pic.count)
      end

   produce_digit(pat : CHARACTER) : CHARACTER is
      do
         if blank_when_zero and then not is_last_char and then cur_num = 0 then
            Result := ' '
         else
            Result := digits.item((cur_num\\10)+1)
            cur_num := cur_num // 10
         end
         is_last_char := False
      end

   produce_blank(pat : CHARACTER) : CHARACTER is
      do
         if sign_leading then
            fill_char  := ifc(pat = 'Z', ' ', '*')
            Result := produce_sign('+')
         elseif not is_last_char and cur_num = 0 then
            Result := ifc(pat = 'Z', ' ', '*')
            is_last_char := False
         else
            Result := produce_digit(pat)
         end
      end

   produce_sign(pat : CHARACTER) : CHARACTER is
      do
         if cur_pic.count > 1 and then
            not contain("+-",cur_pic.item(cur_pic.count-1)) and then
            is_last_char then
               Result := select_sign(pat)
               is_last_char := True
               is_yet_sign := True
         elseif not is_last_char and cur_num = 0 then
            if not is_yet_sign then
               is_yet_sign := True
               Result := select_sign(pat)
            else
               Result := fill_char
            end
         else
            Result := produce_digit('9')
         end
      end

   select_sign(pic_ch : CHARACTER) : CHARACTER is
      do
         if sign = -1 then
            Result := '-'
         else
            Result := ifc(pic_ch = '+', '+', ' ')
         end
      end

   produce_dollar : CHARACTER is
      do
         Result := currency
      end

   produce_noise( pic_ch : CHARACTER) : CHARACTER is
      do
         Result := ifc(pic_ch = 'B', ' ', pic_ch)
         is_last_char := True
      end

   -- Group of procedures for checking of correct conditions

   table : ARRAY[BOOLEAN]

   T_DIGIT       : INTEGER is 1    -- 9
   T_BLANK_ZERO  : INTEGER is 2    -- ' ' and Z
   T_SIGN        : INTEGER is 3    -- + -
   T_DOLLAR      : INTEGER is 4    -- $

   check_digit is
   -- check correct conditions for digit appearance
      do
         if table.item(T_DIGIT) then
            raise_exception("Illegal place for digit")
         end
      end

   check_blank is
   -- check correct conditions for blank appearance
   -- one simple rule: no digits before blank
      do
         if table.item(T_BLANK_ZERO) then
            raise_exception("Illegal place for blank")
         end
         -- no digits before blank
         table.put(True,T_DIGIT)
      end

   check_sign is
   -- check correct conditions for sign appearance
   --  it can be last char
   --  it can be first char
   --  it can be continuous from start optionally after dollar

      do
         if table.item(T_SIGN) then
            raise_exception("Illegal place for sign")
         end
         if cur_pic.count > 1 then
            if is_last_char then
               if contain("+-",cur_pic.item(cur_pic.count-1)) then
               -- all line must be of +- (exception - $)
                  table.put(True,T_DIGIT)
                  table.put(True,T_BLANK_ZERO)
               else
               -- only in last position
                  table.put(True,T_SIGN)
                  is_last_char := True
               end
            else
            -- in middle, at 1st position
               table.put(True,T_DIGIT)
               table.put(True,T_BLANK_ZERO)
            end
         else
            -- 1st position
            table.put(True,T_DIGIT)
            table.put(True,T_BLANK_ZERO)
            table.put(True,T_SIGN)
            table.put(True,T_DOLLAR)
         end
      end

   check_dollar is
   -- check correct conditions for digit appearance
      do
         if table.item(T_DOLLAR) then
            raise_exception("Illegal place for dollar")
         end
         if not (is_last_char or cur_pic.count = 1) then
            raise_exception("Illegal place for dollar")
         end
      end

end -- class INTEGER_COMMERCIAL_FORMATTER
