indexing

   title:       "SCIENTIFIC_FORMAT";
   cluster:     "format/science [format/science]";
   project:     "The Universal Simple Formatting Library (USFL)";
   copyright:   "SIG Computer, 1995";
   author:      "Andy Malakov  (andy@eiffel.ru) and",
                "Igor Nesterof (nest@eiffel.ru)";
   original:    08,Sep,1995;
   version:     1.0;
   last_change: 14,Sep,1995;
   done_at:     "SIG Computer (info@eiffel.ru)";
   extrnl_name: "s_format.e"

class SCIENTIFIC_FORMAT

   creation
      make

   -------------------------------------------------------------------
   feature

      default_escape : CHARACTER

      Asterisk     : CHARACTER is '*' -- take parameter from arguments array
      Dot          : CHARACTER is '.' -- delimeter width | precision
      Format_flags : STRING is "-^+0 " -- available flags
      Minus        : CHARACTER is '-'
      Plus         : CHARACTER is '+'
      Zero         : CHARACTER is '0'
      Space        : CHARACTER is ' '
      Center       : CHARACTER is '^'

      ----------------------------------------------------------------
      make, frozen scientific_make is
         do
            default_escape := '!'
         end -- make

      ----------------------------------------------------------------
      set_escape ( ch : CHARACTER ) is
         do
            default_escape := ch
         end -- set_escape

   --------------------------------------------------------------------
   feature -- Print group of argruments using specified format
   --------------------------------------------------------------------
      printf ( format : STRING; para : ARRAY [ ANY ] ) is
            -- formats list of items into STD_FILES
         require
            Non_void_format : format /= Void
         do
            std_out.put_string( sprintf ( format, para ) )
         end

      ----------------------------------------------------------------
      sprintf ( format : STRING; para : ARRAY [ ANY ] ) : STRING is
            -- formats list of items into string
            -- errors checked only in precondition
         require
            Non_void_format : format /= Void
         do
            from
               Result := first (format)
               !!args.make (para)
            until
               finish
            loop
               str_tools.x_append (Result,argument)
               str_tools.x_append (Result,forth)
               args.go_forth
            end
            args.check_finished

         ensure
            Result /= Void
         end -- sprintf

      ----------------------------------------------------------------
      to_string ( format : STRING; para : ANY ) : STRING is
            -- format single argument to string
         require
            Non_void_paramenters : format /= Void and para /= Void
         do
            Result := sprintf ( format, << para >>)
         ensure
            Result /= Void
         end -- to_string

   -------------------------------------------------------------------
   feature -- Print single element using default format
   -------------------------------------------------------------------
      print ( para : ANY ) is
            -- print argument using default format string
         require
            Non_void_paramenter : para /= Void
         do
            std_out.put_string ( sprint ( para ) )
         end
      ----------------------------------------------------------------
      sprint, frozen typechar_sprint ( para : ANY ) : STRING is
            -- print argument to string using default format string
         require
            Non_void_paramenter : para /= Void
         local
            tfmt : TYPECHAR_FORMATTER
         do
            !!args.make ( << para >> )
            formatters.dyna_types.search_by_example (para)

            tfmt ?= formatters.dyna_types.found_item

            if formatters.dyna_types.found and then tfmt /= Void then
               Result := internal_format ( tfmt.default_format, tfmt, false )
            else
               raise_exception ("Don't know how to print argument")
            end
         ensure
            Result /= Void
         end -- sprint

      ----------------------------------------------------------------
      set_default_format, frozen set_default_typechar_format
         ( format : STRING ) is
            -- Get typechar from format string.
            -- Set this format for all types that conforms
            -- to determined by typechar.
            -- e.g. After call set_default_format( "012x" )
            --      result of print(i) will be the same as printf( "!012x", i)
            --
            -- default escape is optional in format string("!12x" same as "12x")
            --
         require
            valid_format : format /= Void and then
                           format.count > 0 and then
            ( format @ 1 = default_escape implies format.count > 1 )

         local
            typechar : CHARACTER
            format_without_typechar : STRING
         do
            -- Get typechar from last symbol of format string (it must be known)
            typechar := format @ format.count

            -- Search element with this typechar
            formatters.search_by_typechar ( typechar )

            if formatters.found then  -- it's valid typechar
               if format.count > 1 then
                  format_without_typechar := format.substring (1,format.count-1)
               else
                  format_without_typechar := "" -- reset format
               end

               formatters.found_item.
                  set_default_format(format_without_typechar)

               formatters.dyna_types.make_default ( formatters.found_item )

            else
               raise_exception
               ("Unknown format specification after default_escape")
            end
         end -- set_default_format

      ----------------------------------------------------------------
      default_format, frozen default_typechar_format
         ( sample : ANY ) : STRING is
         require
            sample /= Void
         do
            -- Search element with conformed type
            formatters.dyna_types.search_by_example ( sample )
            if formatters.dyna_types.found then
               Result := formatters.dyna_types.found_item.default_format
            end
         end -- default_format

   -------------------------------------------------------------------
   feature { NONE }
   -------------------------------------------------------------------

      ----------------------------------------------------------------
      internal_format ( in : STRING;
                        fmt_el : TYPECHAR_FORMATTER;
                        asterisk_allowed  : BOOLEAN ) : STRING is
            -- Parse format string and interpret format specification
            -- formatspec syntax : [!][flags][width] [.precision]
            -- This feature is private and can't be used alone!
            --
         require
            Non_void_paramenters : in /= Void and
                  fmt_el /= Void and args /= Void
         local
            i, j, k, width, precision : INTEGER
            int_ref : INTEGER_REF
            substr, parsed_flags, set_flags : STRING
            found, was_flags, was_width, was_precision : BOOLEAN
         do

            --debug ("FORMAT")
            --   std_out.put_string ( "FORMAT (" )
            --   std_out.put_string ( in )
            --   std_out.put_string ( ")%N" )
            --end -- debug
            -- -----------------------------------------
            -- skip default_escape (if presents)
            -- -----------------------------------------
            if in.count > 0 and then in @ 1 = default_escape then
                                           k := 2 else k := 1 end

            -- -----------------------------------------
            -- parse flags
            -- -----------------------------------------
--            !!flags.make (0)
--            flags := Format_flags
            !!parsed_flags.make (0)
            !!set_flags.make (0)
            from
               i := k  found := false
            until
               i > in.count or found
            loop
               if str_tools.contain ( Format_flags, in @ i ) and then
                  not ( in @ i = Zero and str_tools.contain ( parsed_flags, in @ i )) then
                  -- if this is flag char but not second zero...
                  was_flags := true
                  set_flag ( in @ i, Void, parsed_flags )
                  i := i + 1
               else
                  found := true -- it's not a flag char
               end
            end -- loop
            -- i points to begin of [width] [.precision]

            -- -----------------------------------------
            -- parse width
            -- -----------------------------------------
            k := str_tools.index ( in, Dot  )
            if k = 0 then -- '.' not found (no precision specified)
               found := true
               k := in.count
            else
               found := false
               k := k - 1
            end
            -- now k points to end of width

            if i <= k then
               substr := in.substring ( i, k )  -- width
               was_width := true

               if asterisk_allowed and then substr @ 1 = Asterisk and then
                  substr.count = 1 then
                  -- take width from arguments list

                  int_ref ?= args.current_item
                  args.go_forth
                  if int_ref /= Void then
                     width := int_ref.item
                  else
                     raise_exception ("Width must conform to INTEGER")
                  end
               else
                  from j := 1 until j > substr.count loop
                     if not str_tools.is_digit ( substr, j ) then
                        raise_exception ("Bad width (non digit character etc)")
                     end
                     j := j + 1
                  end
                  width := str_tools.x_s2i ( substr )
               end
            end

            -- -----------------------------------------
            -- parse precision
            -- -----------------------------------------
            if not found then -- k+1 points to '.'
               if k + 1 < in.count then
                  substr := in.substring (k + 2, in.count )
                  was_precision := true

                  if substr @ 1 = Asterisk and then substr.count = 1 then
                     -- take precision from arguments list
                     int_ref ?= args.current_item
                     args.go_forth
                     if int_ref /= Void then
                        precision := int_ref.item
                     else
                        raise_exception ("Precision must conform to INTEGER")
                     end
                  else
                     from j := 1 until j > substr.count loop
                        if not str_tools.is_digit ( substr, j ) then
                           raise_exception
                              ("Bad precision (non digit character etc)")
                        end
                        j := j + 1
                     end
                     precision := str_tools.x_s2i ( substr )
                  end
               else
                  raise_exception ("Precision missed (after '.')")
               end
            end
            -- -----------------------------------------
            -- interpret format
            -- -----------------------------------------

            if args.current_item = Void then
               raise_exception ("Argument is Void")
            end
            if not args.current_item.conforms_to ( fmt_el.sample ) then
               raise_exception
                  ("Argument does not conform to format specification")
            end

            fmt_el.fmt.make ( args.current_item ) -- initialize fortmat object

            if was_flags then
               from i := 1 until i > parsed_flags.count loop
                  set_flag ( parsed_flags @ i, fmt_el.fmt, set_flags )
                  i := i + 1
               end
            end
            if was_width then
               fmt_el.fmt.set_width ( width )
            end
            if was_precision then
               fmt_el.fmt.set_precision ( precision )
            end

            -- -----------------------------------------
            -- output
            -- -----------------------------------------
            Result := fmt_el.fmt.output

            --debug ("FORMAT")
            --   if was_flags then
            --      std_out.put_string ( " FLAGS (" )
            --      std_out.put_string ( parsed_flags )
            --      std_out.put_string ( ")" )
            --   end
            --   if was_width then
            --      std_out.put_string ( " WIDTH (" )
            --      std_out.put_integer ( width )
            --      std_out.put_string ( ")" )
            --   end
            --   if was_precision then
            --      std_out.put_string ( " PRECISION (" )
            --      std_out.put_integer ( precision )
            --      std_out.put_string ( ")" )
            --   end
            --   std_out.put_string ( "%N" )
            --   std_out.put_string ( "RESULT (")
            --   std_out.put_string ( Result )
            --   std_out.put_string ( ")%N" )
            --end -- debug

         end  -- internal_format

      ----------------------------------------------------------------
      set_flag ( c : CHARACTER, fmt : ABSTRACT_FORMATTER,
                 parsed_flags : STRING ) is
         do

            if not str_tools.contain ( parsed_flags, c ) then
               inspect c
                  when Minus then
                     if str_tools.contain ( parsed_flags, Zero ) then
                        raise_exception ("Flag '-' incompatible with '0'")
                     end
                     if str_tools.contain ( parsed_flags, Center ) then
                        raise_exception ("Flag '-' incompatible with '^'")
                     end
                     if fmt /= Void then fmt.set_align_left () end
                  when Center then
                     if str_tools.contain ( parsed_flags, Zero ) then
                        raise_exception ("Flag '^' incompatible with '0'")
                     end
                     if str_tools.contain ( parsed_flags, Minus ) then
                        raise_exception ("Flag '^' incompatible with '-'")
                     end
                     if fmt /= Void then fmt.set_align_center () end
                  when Plus then
                     if fmt /= Void then fmt.set_sign ( c ) end
                  when Zero then
                     if str_tools.contain ( parsed_flags, Minus ) then
                        raise_exception ("Flag '0' incompatible with '-'")
                     end
                     if str_tools.contain ( parsed_flags, Center ) then
                        raise_exception ("Flag '0' incompatible with '^'")
                     end
                     if fmt /= Void then fmt.set_align_char ( c ) end
                  when Space then
                     if str_tools.contain ( parsed_flags, Plus ) then
                        raise_exception ("Flag ' ' incompatible with '+'")
                     else
                        if fmt /= Void then fmt.set_sign ( c ) end
                     end
               else
                  raise_exception ("Invalid flag")
               end -- inspect
               str_tools.x_extend ( parsed_flags, c )
            else
               raise_exception ("Flag already applied")
            end

         end -- set_flag

      ----------------------------------------------------------------
      formatters, frozen std_formatters :
                         TYPECHAR_FORMAT_TABLE [ TYPECHAR_FORMATTER ] is
         local
            fmt  : TYPECHAR_FORMATTER
            type : ABSTRACT_FORMATTER
            time : TIME
            date : DATE
         once
            !!Result.make

            -- Attention: order of formatters in FORMAT_TABLE
            -- affect result of FORMAT_TABLE.search_by_example

            !INTEGER_FORMATTER!type.make (1)
            !!fmt.make ( 'i', 1, type, "1.1" )
            Result.add ( fmt )

            !!fmt.make ( 'd', 1, type, "1.1" )
            Result.add ( fmt )

            !HEX_FORMATTER!type.make (1)
            !!fmt.make ( 'x', 1, type, "1.1" )
            Result.add ( fmt )

            !OCTAL_FORMATTER!type.make (1)
            !!fmt.make ( 'o', 1, type, "1.1" )
            Result.add ( fmt )

            !UNSIGNED_DECIMAL_INTEGER_FORMATTER!type.make (1)
            !!fmt.make ( 'u', 1, type, "1.1" )
            Result.add ( fmt )

            !CHARACTER_FORMATTER!type.make ('a')
            !!fmt.make ( 'c', 'a', type, "1.1" )
            Result.add ( fmt )

            !BOOLEAN_FORMATTER!type.make ( true )
            !!fmt.make ( 'b', true, type, "5.5" )
            Result.add ( fmt )

            !STRING_FORMATTER!type.make ("a")
            !!fmt.make ( 's',"",type, "" )
            Result.add ( fmt )

            !FLOAT_POINT_FORMATTER!type.make (3.14)
            !!fmt.make ( 'e', 3.14, type, "1.6" )
            Result.add ( fmt )

            !FIXED_POINT_FORMATTER!type.make (3.14)
            !!fmt.make ( 'f', 3.14, type, "1.6" )
            Result.add ( fmt )

            !REAL_FORMATTER!type.make (3.14)
            !!fmt.make ( 'g', 3.14, type, "1.6" )
            Result.add ( fmt )

            !!time.set_to_now
            !TIME_FORMATTER!type.make (time)
            !!fmt.make ( 't', time, type, "" )
            Result.add ( fmt )

            !!date.set_to_now
            !DATE_FORMATTER!type.make (date)
            !!fmt.make ( 'a', date, type, "" )
            Result.add ( fmt )

         end -- formatters

   feature -- secret

      args : ARGUMENTS_ITERATOR

   -------------------------------------------------------------------
   feature {NONE} -- class FORMAT_PARSER
   -------------------------------------------------------------------

      format_str : STRING
      cursor : INTEGER

      ----------------------------------------------------------------
      first ( format : STRING ) : STRING is
         require
            format /= Void
         do
            format_str := format
            cursor := 1
            Result := forth
         end
      ----------------------------------------------------------------
      finish : BOOLEAN is
         do
            Result := cursor > format_str.count
         end
      ----------------------------------------------------------------
      forth : STRING is
         local
            yet_another_argument : BOOLEAN
         do
            !!Result.make (0)
            from
            until
               cursor > format_str.count or yet_another_argument
            loop
               if format_str @ cursor = default_escape then
                  -- check for double default_escape (!!)
                  if cursor < format_str.count and then
                     format_str @ (cursor+1) = default_escape then
                     str_tools.x_extend (Result,default_escape) -- pass this char
                     cursor := cursor + 1 -- skip it
                  else
                     yet_another_argument := true
                  end
               else
                  str_tools.x_extend (Result,format_str @ cursor) -- pass this char
               end
               cursor := cursor + 1 -- go ahead
            end -- loop i

         ensure
            Result /= Void and ( cursor > format_str.count or else
                                 format_str @ (cursor - 1) = default_escape )
         end -- forth

      ----------------------------------------------------------------
      argument, frozen posix_argument : STRING is
            -- parse escape sequence
         require
            valid_format_str : format_str /= Void
            inside_escape : cursor > 1 and then cursor <= format_str.count
                and then format_str @ (cursor - 1) = default_escape
         local
            i : INTEGER
            end_of_escape_seq : BOOLEAN
         do
            !!Result.make (0)
            from
               i := cursor
            until
               i > format_str.count or end_of_escape_seq
            loop
               formatters.search_by_typechar (format_str @ i)
               if formatters.found then  -- it's valid typechar
                  end_of_escape_seq := true
               elseif format_str @ i = default_escape then
                  raise_exception
                  ("Unknown format specification after default_escape")
               end
               i := i + 1
            end -- from i

            if end_of_escape_seq then
               str_tools.x_append ( Result,
                  internal_format (str_tools.x_substring (format_str,cursor,i-2),
                                    formatters.found_item,
                                    true) )
               cursor := i
            else
               raise_exception
               ("Unknown format specification after default_escape")
            end
         ensure
            Result /= Void
         end -- argument

   -------------------------------------------------------------------
   feature { NONE }
   -------------------------------------------------------------------

      is_posix_spec_char ( c : CHARACTER ) : BOOLEAN is
         do
            Result :=
            str_tools.contain ( Format_flags, c ) or else
            ( c >= '0' and c <= '9' )   or else
            c = Asterisk                or else
            c = Dot
         end

   -------------------------------------------------------------------
   feature { NONE }
   -------------------------------------------------------------------
      exc : ERROR_HANDLER is once !!Result end
      raise_exception ( msg : STRING ) is
         local
            m : STRING
         do
            m := "Bad format ("
            --m.append (msg)
            str_tools.x_append (m,msg)
            --m.extend (')')
            str_tools.x_extend (m,')')
            exc.raise ( "SCIENTIFIC_FORMAT", m, 0, Current )
         end -- raise_exception

      ----------------------------------------------------------------
      std_out : STD_FILES is once !!Result end
      str_tools : STRING_TOOLS is once !!Result end

end -- class SCIENTIFIC_FORMAT

